@@ -24,8 +24,10 @@
 #endif
 
 #include <stdio.h>
+#include <unistd.h>
 #include <string.h>
 #include <sys/vfs.h>
+#include <sys/inotify.h>
 #include <glib.h>
 
 #include "connman.h"
@@ -56,6 +58,50 @@ struct connman_config {
 
 static GHashTable *config_table = NULL;
 
+static int inotify_wd = -1;
+
+static GIOChannel *inotify_channel = NULL;
+static uint inotify_watch = 0;
+
+/* Definition of possible strings in the .config files */
+#define CONFIG_KEY_NAME                "Name"
+#define CONFIG_KEY_DESC                "Description"
+
+#define SERVICE_KEY_TYPE               "Type"
+#define SERVICE_KEY_NAME               "Name"
+#define SERVICE_KEY_SSID               "SSID"
+#define SERVICE_KEY_EAP                "EAP"
+#define SERVICE_KEY_CA_CERT            "CACertFile"
+#define SERVICE_KEY_CL_CERT            "ClientCertFile"
+#define SERVICE_KEY_PRV_KEY            "PrivateKeyFile"
+#define SERVICE_KEY_PRV_KEY_PASS       "PrivateKeyPassphrase"
+#define SERVICE_KEY_PRV_KEY_PASS_TYPE  "PrivateKeyPassphraseType"
+#define SERVICE_KEY_IDENTITY           "Identity"
+#define SERVICE_KEY_PHASE2             "Phase2"
+#define SERVICE_KEY_PASSPHRASE         "Passphrase"
+
+static const char *config_possible_keys[] = {
+	CONFIG_KEY_NAME,
+	CONFIG_KEY_DESC,
+	NULL,
+};
+
+static const char *service_possible_keys[] = {
+	SERVICE_KEY_TYPE,
+	SERVICE_KEY_NAME,
+	SERVICE_KEY_SSID,
+	SERVICE_KEY_EAP,
+	SERVICE_KEY_CA_CERT,
+	SERVICE_KEY_CL_CERT,
+	SERVICE_KEY_PRV_KEY,
+	SERVICE_KEY_PRV_KEY_PASS,
+	SERVICE_KEY_PRV_KEY_PASS_TYPE,
+	SERVICE_KEY_IDENTITY,
+	SERVICE_KEY_PHASE2,
+	SERVICE_KEY_PASSPHRASE,
+	NULL,
+};
+
 static void unregister_config(gpointer data)
 {
 	struct connman_config *config = data;
@@ -92,6 +138,33 @@ static void unregister_service(gpointer data)
 	g_free(service);
 }
 
+static void check_keys(GKeyFile *keyfile, const char *group,
+			const char **possible_keys)
+{
+	char **avail_keys;
+	gsize nb_avail_keys, i, j;
+
+	avail_keys = g_key_file_get_keys(keyfile, group, &nb_avail_keys, NULL);
+	if (avail_keys == NULL)
+		return;
+
+	/*
+	 * For each key in the configuration file,
+	 * verify it is understood by connman
+	 */
+	for (i = 0 ; i < nb_avail_keys; i++) {
+		for (j = 0; possible_keys[j] ; j++)
+			if (g_strcmp0(avail_keys[i], possible_keys[j]) == 0)
+				break;
+
+		if (possible_keys[j] == NULL)
+			connman_warn("Unknown configuration key %s in [%s]",
+					avail_keys[i], group);
+	}
+
+	g_strfreev(avail_keys);
+}
+
 static int load_service(GKeyFile *keyfile, const char *group,
 						struct connman_config *config)
 {
@@ -105,6 +178,9 @@ static int load_service(GKeyFile *keyfile, const char *group,
 	if (strlen(ident) < 1)
 		return -EINVAL;
 
+	/* Verify that provided keys are good */
+	check_keys(keyfile, group, service_possible_keys);
+
 	service = g_hash_table_lookup(config->service_table, ident);
 	if (service == NULL) {
 		service = g_try_new0(struct connman_config_service, 1);
@@ -114,19 +190,20 @@ static int load_service(GKeyFile *keyfile, const char *group,
 		service->ident = g_strdup(ident);
 	}
 
-	str = g_key_file_get_string(keyfile, group, "Type", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_TYPE, NULL);
 	if (str != NULL) {
 		g_free(service->type);
 		service->type = str;
 	}
 
-	str = g_key_file_get_string(keyfile, group, "Name", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_NAME, NULL);
 	if (str != NULL) {
 		g_free(service->name);
 		service->name = str;
 	}
 
-	hex_ssid = g_key_file_get_string(keyfile, group, "SSID", NULL);
+	hex_ssid = g_key_file_get_string(keyfile, group, SERVICE_KEY_SSID,
+					 NULL);
 	if (hex_ssid != NULL) {
 		char *ssid;
 		unsigned int i, j = 0, hex;
@@ -163,57 +240,58 @@ static int load_service(GKeyFile *keyfile, const char *group,
 		service->ssid_len = ssid_len;
 	}
 
-	str = g_key_file_get_string(keyfile, group, "EAP", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_EAP, NULL);
 	if (str != NULL) {
 		g_free(service->eap);
 		service->eap = str;
 	}
 
-	str = g_key_file_get_string(keyfile, group, "CACertFile", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_CA_CERT, NULL);
 	if (str != NULL) {
 		g_free(service->ca_cert_file);
 		service->ca_cert_file = str;
 	}
 
-	str = g_key_file_get_string(keyfile, group, "ClientCertFile", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_CL_CERT, NULL);
 	if (str != NULL) {
 		g_free(service->client_cert_file);
 		service->client_cert_file = str;
 	}
 
-	str = g_key_file_get_string(keyfile, group, "PrivateKeyFile", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_PRV_KEY, NULL);
 	if (str != NULL) {
 		g_free(service->private_key_file);
 		service->private_key_file = str;
 	}
 
 	str = g_key_file_get_string(keyfile, group,
-						"PrivateKeyPassphrase", NULL);
+						SERVICE_KEY_PRV_KEY_PASS, NULL);
 	if (str != NULL) {
 		g_free(service->private_key_passphrase);
 		service->private_key_passphrase = str;
 	}
 
 	str = g_key_file_get_string(keyfile, group,
-					"PrivateKeyPassphraseType", NULL);
+					SERVICE_KEY_PRV_KEY_PASS_TYPE, NULL);
 	if (str != NULL) {
 		g_free(service->private_key_passphrase_type);
 		service->private_key_passphrase_type = str;
 	}
 
-	str = g_key_file_get_string(keyfile, group, "Identity", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_IDENTITY, NULL);
 	if (str != NULL) {
 		g_free(service->identity);
 		service->identity = str;
 	}
 
-	str = g_key_file_get_string(keyfile, group, "Phase2", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_PHASE2, NULL);
 	if (str != NULL) {
 		g_free(service->phase2);
 		service->phase2 = str;
 	}
 
-	str = g_key_file_get_string(keyfile, group, "Passphrase", NULL);
+	str = g_key_file_get_string(keyfile, group, SERVICE_KEY_PASSPHRASE,
+					NULL);
 	if (str != NULL) {
 		g_free(service->passphrase);
 		service->passphrase = str;
@@ -240,13 +318,16 @@ static int load_config(struct connman_config *config)
 	if (keyfile == NULL)
 		return -EIO;
 
-	str = g_key_file_get_string(keyfile, "global", "Name", NULL);
+	/* Verify keys validity of the global section */
+	check_keys(keyfile, "global", config_possible_keys);
+
+	str = g_key_file_get_string(keyfile, "global", CONFIG_KEY_NAME, NULL);
 	if (str != NULL) {
 		g_free(config->name);
 		config->name = str;
 	}
 
-	str = g_key_file_get_string(keyfile, "global", "Description", NULL);
+	str = g_key_file_get_string(keyfile, "global", CONFIG_KEY_DESC, NULL);
 	if (str != NULL) {
 		g_free(config->description);
 		config->description = str;
@@ -266,18 +347,18 @@ static int load_config(struct connman_config *config)
 	return 0;
 }
 
-static int create_config(const char *ident)
+static struct connman_config *create_config(const char *ident)
 {
 	struct connman_config *config;
 
 	DBG("ident %s", ident);
 
 	if (g_hash_table_lookup(config_table, ident) != NULL)
-		return -EEXIST;
+		return NULL;
 
 	config = g_try_new0(struct connman_config, 1);
 	if (config == NULL)
-		return -ENOMEM;
+		return NULL;
 
 	config->ident = g_strdup(ident);
 
@@ -288,9 +369,7 @@ static int create_config(const char *ident)
 
 	connman_info("Adding configuration %s", config->ident);
 
-	load_config(config);
-
-	return 0;
+	return config;
 }
 
 static int read_configs(void)
@@ -320,9 +399,13 @@ static int read_configs(void)
 
 			ident = g_string_free(str, FALSE);
 
-			if (connman_dbus_validate_ident(ident) == TRUE)
-				create_config(ident);
+			if (connman_dbus_validate_ident(ident) == TRUE) {
+				struct connman_config *config;
 
+				config = create_config(ident);
+				if (config != NULL)
+					load_config(config);
+			}
 			g_free(ident);
 		}
 
@@ -332,6 +415,148 @@ static int read_configs(void)
 	return 0;
 }
 
+static gboolean inotify_data(GIOChannel *channel, GIOCondition cond,
+							gpointer user_data)
+{
+	char buffer[256];
+	char *next_event;
+	gsize bytes_read;
+	GIOError err;
+
+	if (cond & (G_IO_NVAL | G_IO_ERR | G_IO_HUP)) {
+		inotify_watch = 0;
+		return FALSE;
+	}
+
+	err = g_io_channel_read(channel, buffer,
+					sizeof(buffer) - 1, &bytes_read);
+
+	if (err != G_IO_ERROR_NONE) {
+		if (err == G_IO_ERROR_AGAIN)
+			return TRUE;
+
+		connman_error("Reading from inotify channel failed");
+		inotify_watch = 0;
+		return FALSE;
+	}
+
+	next_event = buffer;
+
+	while (bytes_read > 0) {
+		struct inotify_event *event;
+		gchar *ext;
+		gchar *ident;
+		gsize len;
+
+		event = (struct inotify_event *) next_event;
+		if (event->len)
+			ident = next_event + sizeof(struct inotify_event);
+		else
+			ident = NULL;
+
+		len = sizeof(struct inotify_event) + event->len;
+
+		/* check if inotify_event block fit */
+		if (len > bytes_read)
+			break;
+
+		next_event += len;
+		bytes_read -= len;
+
+		if (ident == NULL)
+			continue;
+
+		if (g_str_has_suffix(ident, ".config") == FALSE)
+			continue;
+
+		ext = g_strrstr(ident, ".config");
+		if (ext == NULL)
+			continue;
+
+		*ext = '\0';
+
+		if (connman_dbus_validate_ident(ident) == FALSE)
+			continue;
+
+		if (event->mask & IN_CREATE)
+			create_config(ident);
+
+		if (event->mask & IN_MODIFY) {
+			struct connman_config *config;
+
+			config = g_hash_table_lookup(config_table, ident);
+			if (config != NULL) {
+				g_hash_table_remove_all(config->service_table);
+				load_config(config);
+			}
+		}
+
+		if (event->mask & IN_DELETE)
+			g_hash_table_remove(config_table, ident);
+	}
+
+	return TRUE;
+}
+
+static int create_watch(void)
+{
+	int fd;
+
+	fd = inotify_init();
+	if (fd < 0)
+		return -EIO;
+
+	inotify_wd = inotify_add_watch(fd, STORAGEDIR,
+					IN_MODIFY | IN_CREATE | IN_DELETE);
+	if (inotify_wd < 0) {
+		connman_error("Creation of STORAGEDIR  watch failed");
+		close(fd);
+		return -EIO;
+	}
+
+	inotify_channel = g_io_channel_unix_new(fd);
+	if (inotify_channel == NULL) {
+		connman_error("Creation of inotify channel failed");
+		inotify_rm_watch(fd, inotify_wd);
+		inotify_wd = 0;
+
+		close(fd);
+		return -EIO;
+	}
+
+	g_io_channel_set_close_on_unref(inotify_channel, TRUE);
+	g_io_channel_set_encoding(inotify_channel, NULL, NULL);
+	g_io_channel_set_buffered(inotify_channel, FALSE);
+
+	inotify_watch = g_io_add_watch(inotify_channel,
+				G_IO_IN | G_IO_HUP | G_IO_NVAL | G_IO_ERR,
+				inotify_data, NULL);
+
+	return 0;
+}
+
+static void remove_watch(void)
+{
+	int fd;
+
+	if (inotify_channel == NULL)
+		return;
+
+	if (inotify_watch > 0) {
+		g_source_remove(inotify_watch);
+		inotify_watch = 0;
+	}
+
+	fd = g_io_channel_unix_get_fd(inotify_channel);
+
+	if (inotify_wd >= 0) {
+		inotify_rm_watch(fd, inotify_wd);
+		inotify_wd = 0;
+	}
+
+	g_io_channel_unref(inotify_channel);
+}
+
 int __connman_config_init(void)
 {
 	DBG("");
@@ -339,6 +564,8 @@ int __connman_config_init(void)
 	config_table = g_hash_table_new_full(g_str_hash, g_str_equal,
 						NULL, unregister_config);
 
+	create_watch();
+
 	return read_configs();
 }
 
@@ -346,6 +573,8 @@ void __connman_config_cleanup(void)
 {
 	DBG("");
 
+	remove_watch();
+
 	g_hash_table_destroy(config_table);
 	config_table = NULL;
 }
